######################################################################
# main ZWiki source file
"""
ZWiki README

This zope product allows you to build wiki webs with zope objects.
The included ZWikiWebs product provides sample wikiwebs to help you
get started quickly. There is a documentation & discussion site
running at http://zwiki.org .

You are responsible for security; zwiki may contain unforeseen risks.
I appreciate all feedback, bugreports, fixes, responses to questions
in the code, re-implementations, etc. 

(c) 1999,2000,2001 Simon Michael <simon@joyful.com>
This product is available under the GPL - see LICENSE.txt
All rights reserved, all disclaimers apply, etc.

Installation:

- unpack ZWiki-x.x.x.tgz in your zope Products directory, restart zope

- put ZWikiWebs.zexp in your zope import directory, import into
/Control_Panel/Products

- create a ZWiki Web, view it

*** NB! DTML methods and sample wiki pages from earlier versions will
not work with ZWiki 0.7.0 and greater. There have been api
enhancements, zope 2.2 compatibility changes, etc. So if you have an
existing pre-0.7.0 wiki web, you can either 1. delete all your DTML
methods and let zwiki supply defaults or 2. compare your methods with
those from the sample wikis and make just the necessary changes.



TESTS: (see tests/DocTests.py)

"""

__version__="$Revision: 0.72 $"[11:-2]

#'

import re
from string import split,join,find,lower,rfind,atoi
from types import *
from urllib import quote, unquote       # *url* quoting
from wwml import translate_WMML
from AccessControl import getSecurityManager
import Acquisition
from App.Common import absattr, rfc1123_date, aq_base
from DocumentTemplate import HTML
from DocumentTemplate.DT_Util import html_quote # *html* quoting
import Globals
from Globals import HTMLFile, MessageDialog
from Products.MailHost.MailHost import MailBase
from OFS.content_types import guess_content_type
from OFS.Document import Document
from OFS.DTMLDocument import DTMLDocument
import OFS.Image
import StructuredText
#from Products.ZCatalog.CatalogAwareness import CatalogAware # catalog code

from ZWikiDefaults import DEFAULT_PAGE_TYPE, DISABLE_JAVASCRIPT, \
     LARGE_FILE_SIZE, default_wiki_page, default_wiki_header,\
     default_wiki_footer, default_editform, default_backlinks
from ZWikiDiff import ZWikiDiffMixin
from ZWikiParents import ZWikiParentsMixin
import ZWikiPermissions
from ZWikiRegexes import intl_char_entities, urlchars, urlendchar, \
     url, bracketedexpr, wikiname1, wikiname2, simplewikilink, \
     wikilink, interwikilink, remotewikiurl, pagesubscribers, protected_line



class ZWikiPage(DTMLDocument, ZWikiParentsMixin, ZWikiDiffMixin): # CatalogAware
    """
    A ZWikiPage is a DTML Document which knows how to render itself in
    various wiki styles, and provides some utility methods to support
    wiki-building.

    TESTS:

        
    """

    # define various meta-info & attributes
    # I wonder if some of these should be in the __init__ ? do we care ?
    # ah, these are class (vs instance) attributes. the "backwards
    # compatibility" code can probably be simplified.
    meta_type = "ZWiki Page"
    icon      = "misc_/ZWiki/ZWikiPage_icon"
    page_type = DEFAULT_PAGE_TYPE
    last_editor = ''
    last_editor_ip = ''

    # define properties visible in the management interface
    # how would you append to the superclass' _properties here ?
    # ? DocumentTemplate.inheritedAttribute('_properties'),...) ?
    _properties=(
        {'id':'title', 'type': 'string', 'mode':'w'},
        {'id':'page_type', 'type': 'string', 'mode': 'w'},
        {'id':'last_editor', 'type': 'string', 'mode': 'r'},
        {'id':'last_editor_ip', 'type': 'string', 'mode': 'r'},
        #useful ?
        #{'id':'creator', 'type': 'string', 'mode': 'r'},
        #{'id':'creator_ip', 'type': 'string', 'mode': 'r'},
        ) + ZWikiParentsMixin._properties # XXX doable ? better way ?
                                          # parent code

    # define permissions
    # Most of these we check in the code ourselves, because edit()
    # does a lot of different things. This also allows more helpful
    # authorization errors.
    __ac_permissions__=(
        (ZWikiPermissions.Append, ()),
        (ZWikiPermissions.ChangeType, ()),
        (ZWikiPermissions.Change, ('PUT','manage_edit','manage_upload')),
        (ZWikiPermissions.Delete, ()),
        ) + ZWikiParentsMixin.__ac_permissions__ # XXX doable ? better way ?
                                                 # parent code

    def __repr__(self):
        return ("<%s %s at 0x%s>"
                % (self.__class__.__name__, `self.id()`, hex(id(self))[2:]))

    # method for returning this page's raw text
    # - was called src (kept for backwards compatibility)
    # - renamed to text
    # - should it be available as document_src for consistency with zope ?
    # XXX why does this need to be permission-free ?
    text__roles__ = None
    def text(self, REQUEST=None, RESPONSE=None):
        """Permission-free version of document_src"""
        if RESPONSE is not None:
            RESPONSE.setHeader('Content-Type', 'text/plain')
            #RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime))# XXX see above
        return self.xread()

    src__roles__ = None
    src = text
    #document_src__roles__ = None
    #document_src = text

    # XXX see render_structuredtextdtml
    def xread(self):
        return re.sub('<!--antidecapitationkludge-->\n\n?','', self.read())
        # old pages have a single \n, new ones have two
        # in the window of time between changing to two and adding the ?,
        # pages acquired two of these markers when edited. The extra
        # should get removed at the next edit.


    ########################################
    # methods for page-rendering
    
    def __call__(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """Render a zwiki page, with standard header & footer
        """
        # backwards compatibility - fix up old zwikipages
        # XXX Note that the objects don't get very far unpickling, some
        #     by-hand adjustment via command-line interaction is necessary to
        #     get them over the transition, sigh. --ken
        #     When does this come about ? --SM

        # Confused about what happens in the zodb when class
        # definitions change. I now think that all instances in the
        # zodb conform to the new class shape immediately on
        # refresh/restart. True ? Not sure what exactly happens to
        # deleted properties/attributes - discarded ? No I think they
        # lurk around.
        
        # XXX should all this be done offline by a script ?
        if (not hasattr(self, 'parents')) or self.parents is None:
            self.parents = []           # XXX parent code
        if not hasattr(self, 'last_editor'):
            self.last_editor = ''
        if not hasattr(self, 'last_editor_ip'):
            self.last_editor_ip = ''

        if (hasattr(self, 'username') and
            type(self.username) is StringType): 
            if re.match(r'[0-9\.]+',self.username):
                self.last_editor_ip = self.username
            else:
                self.last_editor = self.username
            delattr(self,'username')
            self.last_editor_ip = ''
            self.last_editor = 'UpGrade'
            
            get_transaction().commit()
            # a bit confusing - here's what I believe I'm doing:
            # when we encounter a zwikipage with a username string
            # attribute (left over from the old username property), we
            # copy the info to the appropriate last_editor* property
            # and delete the old attribute. Once it's gone, the
            # username method below is revealed and allows old dtml to
            # work as before - handy. Also, do a commit to update the
            # page's timestamp immediately, otherwise the append form
            # will get displayed with an out of date timestamp.

        if not hasattr(self, 'page_type'):
            self.page_type = DEFAULT_PAGE_TYPE
        elif self.page_type == 'Structured Text':
            self.page_type = 'structuredtext'
        elif self.page_type == 'structuredtext_dtml':
            self.page_type = 'structuredtextdtml'
        elif self.page_type[0:4] == 'HTML':
            self.page_type = 'html'
        elif  self.page_type == 'html_dtml':
            self.page_type = 'htmldtml'
        elif self.page_type == 'Classic Wiki':
            self.page_type = 'classicwiki'
        elif self.page_type == 'Plain Text':
            self.page_type = 'plaintext'

        # prepend wiki header
        # XXX experiments to see if we can allow a zwikipage to work
        # as well as a dtml method
        # actually a flag in REQUEST is not much use, try a keyword arg
        if (REQUEST is not None and
            not (hasattr(REQUEST,'noheaders') or
                 hasattr(REQUEST,'bare') or
                 hasattr(REQUEST,'noheader') or
                 kw.has_key('bare') or
                 kw.has_key('noheader'))):
            if hasattr(self, 'standard_wiki_header'):
                if self.standard_wiki_header.meta_type is "ZWiki Page":
                    #REQUEST['noheader']=1
                    REQUEST.noheader=1
                hdr = apply(self.standard_wiki_header.__call__,
                            (self, REQUEST, RESPONSE), kw)
                if self.standard_wiki_header.meta_type is "ZWiki Page":
                    hdr=unquote(hdr)
                    hdr=re.sub(r'</?pre>',r'',hdr)
            else:
                hdr = default_wiki_header(self,REQUEST)
        else: hdr = ''

        # call render_<page_type> to render the body
        r = 'render_' + self.page_type
        if hasattr(self, r):
            r = getattr(self,r)
        else:
            r = self.render_plaintext

        body = apply(r,(self, REQUEST, RESPONSE), kw)

        # append wiki footer
        if (REQUEST is not None and
            not (hasattr(REQUEST,'noheaders') or
                 hasattr(REQUEST,'bare') or
                 hasattr(REQUEST,'nofooter') or
                 kw.has_key('bare') or
                 kw.has_key('nofooter'))):
            if hasattr(self, 'standard_wiki_footer'):
                if self.standard_wiki_footer.meta_type is "ZWiki Page":
                    REQUEST.nofooter=1
                ftr = apply(self.standard_wiki_footer.__call__,
                            (self, REQUEST, RESPONSE), kw)
                if self.standard_wiki_footer.meta_type is "ZWiki Page":
                    ftr=unquote(ftr)
                    ftr=re.sub(r'</?pre>',r'',ftr)
            else:
                ftr = default_wiki_footer(self,REQUEST)
        else: ftr = ''

	if RESPONSE is not None:
	    RESPONSE.setHeader('Content-Type', 'text/html')
            #RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime)) 
            # XXX causes browser caching strangeness, why ? 

        return hdr + body + ftr

    # built-in render methods; add new ones with dtml/python/external
    # methods. Their names correspond with the page_type attribute. NB
    # if the name contains "dtml", we will do DTML validation after
    # each edit.
    #
    # I inlined these for maximum control

    def render_structuredtextdtml(self, client=None, REQUEST={},
                                  RESPONSE=None, **kw): 
        # DTML + structured text + wiki links + HTML

        t = apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)
        # XXX problem: an interesting DTML "feature".  DTML
        # decapitates() initial lines that look like http headers,
        # moving them into the actual HTTP header.
        # (not-at-all-)quick workaround: keep a comment in front to hide them,
        # omit it everywhere else; see _set_text, xread

        t = re.sub(protected_line, self._protect_line, t)
        t = self._structured_text(t)
        t = re.sub(interwikilink,
                   thunk_substituter(self._interwikilink_replace, t, 1),
                   t)
        t = re.sub(wikilink,
                   thunk_substituter(self._wikilink_replace, t, 1),
                   t)
        return t

    def render_structuredtext(self, client=None, REQUEST={},
                              RESPONSE=None, **kw):
        # structured text + wiki links + HTML

        t = str(self.xread())
        t = re.sub(protected_line, self._protect_line, t)
        t = self._structured_text(t)
        t = re.sub(interwikilink,
                   thunk_substituter(self._interwikilink_replace, t, 1),
                   t)
        t = re.sub(wikilink,
                   thunk_substituter(self._wikilink_replace, t, 1),
                   t)
        return t

    #XXX this should be structuredtext and the above should be structuredtexthtml
    def render_structuredtextonly(self, client=None, REQUEST={},
                              RESPONSE=None, **kw):
        # structured text + wiki links

        t = html_quote(self.xread())
        t = re.sub(protected_line, self._protect_line, t)
        t = self._structured_text(t)
        t = re.sub(interwikilink,
                   thunk_substituter(self._interwikilink_replace, t, 1),
                   t)
        t = re.sub(wikilink,
                   thunk_substituter(self._wikilink_replace, t, 1),
                   t)
        return t

    def render_htmldtml(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        # DTML + wiki links + HTML

        t = apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)

        t = re.sub(protected_line, self._protect_line, t)
        t = re.sub(interwikilink,
                   thunk_substituter(self._interwikilink_replace, t, 1),
                   t)
        t = re.sub(wikilink,
                   thunk_substituter(self._wikilink_replace, t, 1),
                   t)
        return t

    def render_html(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        # wiki links + HTML
        t = str(self.xread())
        t = re.sub(protected_line, self._protect_line, t)
        t = re.sub(interwikilink,
                   thunk_substituter(self._interwikilink_replace, t, 1),
                   t)
        t = re.sub(wikilink,
                   thunk_substituter(self._wikilink_replace, t, 1),
                   t)
        return t

    def render_plainhtml(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        # HTML with no surprises
        t = str(self.xread())
        return t

    def render_plainhtmldtml(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        # DTML + HTML, nothing else
        t = apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)
        return t

    def render_classicwiki(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        # classic wiki formatting + (inter)wiki links
        t = html_quote(self.xread())
        t = translate_WMML(t)
        t = re.sub(interwikilink,
                   thunk_substituter(self._interwikilink_replace, t, 1),
                   t)
        t = re.sub(simplewikilink,
                   thunk_substituter(self._wikilink_replace, t, 1),
                   t)
        return t

    def render_plaintext(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        # fixed-width plain text with no surprises
        t = "<pre>\n" + html_quote(self.xread()) + "\n</pre>\n"
        return t

    # NB! StructuredTextNG has been quietly added to zope.
    # So we have some greatly expanded possibilities here

    #XXX this could be a function. should it be ?
    def _structured_text(self, text):
        """
        Render some text into html according to the structured text
        rules, massaging and tweaking slightly to work around stx
        issues and for our nefarious zwiki purposes.
        """
        # strip trailing blank lines to avoid the famous last line
        # heading bug
        text = re.sub(r'(?m)\n[\n\s]*$', r'\n', text)

        # "A paragraph that begins with a sequence of sequences, where
        # each sequence is a sequence of digits or a sequence of
        # letters followed by a period, is treated as an ordered list
        # element."
        # This stx rule converts an initial single word plus period
        # into a numeric bullet. Don't understand why it exists.
        # Quickest workaround: prepend a harmless marker like <!---->
        # preserve indentation
        # now why did I need (?m) here ? seems backwards
        # hmm \s matches our line breaks
        # caution - our workaround will be visible eg in :: examples,
        # so we remove it after stx has done it's thing, below
        text = re.sub(r'(?m)^([ \t]*)([A-z]\w*\.)', r'\1<!--STXKLUDGE-->\2', text)
     
        # "Sub-paragraphs of a paragraph that ends in the word example
        # or the word examples, or :: is treated as example code and
        # is output as is"
        # This fails if there is whitespace after the ::, so remove it
        # NB contrary to the docs "the word example or examples" is
        # not used, rightly I think
        text = re.sub(r'(?m)::[ \t]+$', r'::', text)

        # skip stx's [] footnote linking
        
        text = str(StructuredText.HTML(text,level=3))

        # oh god the things I do in the name of DTSTTCPW
        text = re.sub(r'(?i)(<|&lt;)!--STXKLUDGE--(>|&gt;)', r'', text)

        return text

    def wikilink(self, text):
        """
        utility method for wiki-linking an arbitrary text snippet

        TESTS:

        >>> zc.TestPage.wikilink('http://a.b.c/d')
        '<a href="http://a.b.c/d">http://a.b.c/d</a>'

        >>> zc.TestPage.wikilink('mailto://a@b.c')
        '<a href="mailto://a@b.c">mailto://a@b.c</a>'

        >>> zc.TestPage.wikilink('nolink')
        'nolink'

        >>> zc.TestPage.wikilink('TestPage')[-23:]
        '/TestPage">TestPage</a>'

        >>> zc.TestPage.wikilink('TestPage1234')[-43:]
        '/TestPage/editform?page=TestPage1234">?</a>'

        >>> zc.TestPage.wikilink('!TestPage')
        'TestPage'

        >>> zc.TestPage.wikilink('[nolink]')[-37:]
        '/TestPage/editform?page=nolink">?</a>'

        a problem with escaping remote wiki links was reported
        >>> zc.TestPage.edit(text='RemoteWikiURL: URL/')
        >>> zc.TestPage.wikilink('TestPage:REMOTEPAGE') # should give:
        '<a href="URL/REMOTEPAGE">TestPage:REMOTEPAGE</a>'
        >>> zc.TestPage.wikilink('!TestPage:REMOTEPAGE') # should give:
        'TestPage:REMOTEPAGE'

        """
        # a ' for emacs
            
        # could modify the render_ methods to do this instead
        t = re.sub(protected_line, self._protect_line, text)
        t = re.sub(interwikilink,
                   thunk_substituter(self._interwikilink_replace, t, 1),
                   t)
        t = re.sub(wikilink,
                   thunk_substituter(self._wikilink_replace, t, 1),
                   t)
        return t

    # expose this darn thing to dtml python expressions once and for all!
    def htmlquote(self, text):
        return html_quote(text)

    # private workhorse methods for text-crunching & wikification
    # tasty spaghetti regexps! better suggestions welcome ?

    def _protect_line(self, match):
        """protect an entire line from _wikilink_replace, by
           protecting all its wikilink occurrences
        """
        return re.sub(wikilink, r'!\1', match.group(1))

    def _wikilink_replace(self, match, allowed=0, state=None, text=''):
        """
        Replace an occurrence of the wikilink regexp or one of the
        special [] constructs with a suitable hyperlink

        To be used as a re.sub repl function *and* get a proper value
        for literal context, 'allowed', etc, enclose this function
        with the value using 'thunk_substituter'.
        """
        # In a literal?
        if state is not None:
            if within_literal(match.start(1), match.end(1)-1, state, text):
                return match.group(1)

        # matches beginning with ! should be left alone
        if re.match('^!',match.group(0)):
            return match.group(1)

        m = match.group(1)

        # if it's a bracketed expression,
        if re.match(bracketedexpr,m):

            # strip the enclosing []'s
            m = re.sub(bracketedexpr, r'\1', m)

            # extract a (non-url) path if there is one
            pathmatch = re.match(r'(([^/]*/)+)([^/]+)',m)
            if pathmatch:
                path,id = pathmatch.group(1), pathmatch.group(3)
            else:
                path,id = '',m

            # if it looks like an image link, inline it
            if guess_content_type(id)[0][0:5] == 'image':
                return '<img src="%s%s">' % (path,id)

            # or if there was a path assume it's to some non-wiki
            # object and skip the usual existence checking for
            # simplicity. Could also attempt to navigate the path in
            # zodb to learn more about the destination
            if path:
                return '<a href="%s%s">%s</a>' % (path,id,id)

            # otherwise fall through to normal link processing
        
        # if it's an ordinary url, link to it
        if re.match(url,m):
            # except, if preceded by " or = it should probably be left alone
            if re.match('^["=]',m):                                   # "
                return m
            else:
                return '<a href="%s">%s</a>' % (m, m)

        # a wikiname - if a page (or something) of this name exists, link to it
        elif hasattr(self.aq_parent, m):
            #return '<a href="%s">%s</a>' % (quote(m), m)
            # XXX make all wiki links absolute. this is a bit drastic!!
            # It's to make eg BookMarks more robust in editform & backlinks,
            # hopefully it won't break anything. I have no better ideas
            # right now so it's Do The Simplest Thing time
            return '<a href="%s/%s">%s</a>' % (self.wiki_url(),quote(m), m)

        # otherwise, provide a "?" creation link
        else:
            # XXX see above
            #return '%s<a href="%s/editform?page=%s">?</a>' % (m, quote(self.id()), quote(m))
            return '%s<a href="%s/%s/editform?page=%s">?</a>' % (m, self.wiki_url(), quote(self.id()), quote(m))
        
    def _interwikilink_replace(self, match, allowed=0, state=None, text=''):
        """
        Replace an occurrence of interwikilink with a suitable hyperlink.

        To be used as a re.sub repl function *and* get a proper value
        for literal context, 'allowed', etc, enclose this function
        with the value using 'thunk_substituter'.
        """
        # matches beginning with ! should be left alone
        #if re.match('^!',match.group(0)): return match.group(1)
        # NB this is a bit naughty, but: since we know this text will likely
        # be scanned with _wikilink_replace right after this pass, leave
        # the ! in place for it to find. Otherwise the localname will
        # get wiki-linked.
        if re.match('^!',match.group(0)): return match.group(0)

        localname  = match.group('local')
        remotename = match.group('remote') # named groups come in handy here!

        # NB localname could be [bracketed]
        if re.match(bracketedexpr,localname):
            localname = re.sub(bracketedexpr, r'\1', localname)

        # look for a RemoteWikiURL definition
        if hasattr(self.aq_parent, localname): 
            localpage = getattr(self.aq_parent,localname)
            # local page found - search for "RemoteWikiUrl: url"
            m = re.search(remotewikiurl, str(localpage))
            if m is not None:
                remoteurl = html_unquote(m.group(1)) # NB: pages are 
                                                     # stored html-quoted
                                                     # XXX eh ? they are ?
                                                     # something's not
                                                     # right
                                                     # somewhere.. 
                                                     # I have lost my
                                                     # grip on this
                                                     # whole quoting
                                                     # issue.
                
                # we have a valid inter-wiki link
                link = '<a href="%s%s">%s:%s</a>' % \
                       (remoteurl, remotename, localname, remotename)
                # protect it from any later wiki-izing passes
                return re.sub(wikilink, r'!\1', link)

        # otherwise, leave alone
        return match.group(0)


    ########################################
    # methods for editing & creating pages

    # this code has become more complex to support late page creation,
    # but the api should now be more general & powerful than it was
    # XXX refactor

    # XXX can we simplify ? why shouldn't this logic be in the
    # editform dtml method(s)
    def editform(self, REQUEST=None, page=None, text=None, action='Change'):
        """
        Display a form which will edit or create the specified page.
        For new pages, initial text may be specified.
        The form may be overridden by a DTML method called 'editform'.
        """
        # what are we going to do ? set up page, text & action accordingly
        if page is None:
            # no page specified - editing the current page
            page = self.id()
            text = self.xread()
        elif hasattr(self.aq_parent, page):
            # editing another existing page
            text = getattr(self.aq_parent,page).xread()
        else:
            # editing a new page
            action = 'Create'
            # supply default text if needed
            if text is None:
                if hasattr(self.aq_parent, 'standard_wiki_page'):
                    if callable(self.aq_parent.standard_wiki_page):
                        text = self.aq_parent.standard_wiki_page(self,REQUEST)
                    else:
                        text = HTML(self.aq_parent.standard_wiki_page)
                else:
                    text = default_wiki_page(self,REQUEST)

        # display the edit form - a dtml method or the builtin default
        # NB we redefine id as a convenience, so that one header can work
        # for pages and editforms
        # XXX can we simplify this/make dtml more version independent ?
        if hasattr(self.aq_parent, 'editform'):
            return self.aq_parent.editform(self,REQUEST,
                                           page=page,text=text,action=action,
                                           id=page,oldid=self.id())
        else:
            return default_editform(self,REQUEST,
                                    page=page,text=text,action=action,
                                    id=page,oldid=self.id())

    def append(self, text='', separator='\n\n', REQUEST=None):
        """
        Appends some text to an existing zwiki page.
        Also sends email notifications if appropriate.
        """
        oldtext = self.xread()
        newtext = str(text)

        # XXX quick & dirty email notification on append
        # search for recipients following "PageSubscribers:";
        # if these look vaguely like email addresses and we have a
        # mailhost and a from address, send a notification
        # should be done after edit has succeeded though
        # XXX do this on edit also
        m = re.search(pagesubscribers, oldtext)
        if m:
            subscribers = html_unquote(m.group(1))
            if re.search('@',subscribers):
                # from SendMailTag.py
                if hasattr(self, 'MailHost'):
                    mhost=self.MailHost
                    if hasattr(self.aq_parent,'mail_from'):
                        fromaddr=self.aq_parent.mail_from

                        # by default, append a signature with page url
                        if (hasattr(self.aq_parent, 'mail_no_signature') and
                            self.aq_parent.mail_no_signature):
                            signature = ''
                        else:
                            signature = """\

---
forwarded from %s
""" % (self.wiki_page_url())

                        # send notification
                        # don't touch that \ !
                        mhost.send("""\
From: %s (%s)
To: %s
Subject: %s comment

%s
%s
""" % (fromaddr,self.zwiki_username_or_ip(REQUEST),subscribers,self.id(),
       newtext,signature))


        # usability hack: scroll to bottom after adding a comment
        if REQUEST:
            REQUEST['URL1'] = REQUEST['URL1'] + '#bottom'

        # call edit to do the work
        # NB we do allow a blank comment to send out a notification;
        # don't bother calling edit since it will do nothing.
        # on the other hand, it might be nice to get a record of who did this
        if newtext is not '':
            self.edit(text=oldtext+separator+newtext, REQUEST=REQUEST)
            # don't need to worry about timeStamp since edit conflicts
            # aren't a problem with appends
        

    def edit(self, page=None, text=None, type=None, title='', 
             timeStamp=None, REQUEST=None):
        """
        General-purpose method for editing & creating zwiki pages.
        Changes the text and/or markup type of this (or the specified)
        page, or creates the specified page if it does not exist.

        Other special features:

        - Usually called from a time-stamped web form; we use
        timeStamp to detect and warn when two people attempt to work
        on a page at the same time. This makes sense only if timeStamp
        came from an editform for the page we are actually changing.

        - The username (authenticated user or zwiki_username cookie)
        and ip address are saved in page's last_editor, last_editor_ip
        attributes if a change is made

        - If the text begins with "DeleteMe", move this page to the
        recycle_bin subfolder.

        - If file has been submitted in REQUEST, create a file or
        image object and link or inline it on the current page.

        Doing all this stuff in one method simplifies the layer above
        I think.

        XXX refactor 

        TESTS:

        >>> # set a page's text to something
        >>> zc.TestPage.edit(text='something')
        >>> zc.TestPage.src()
        'something'
        >>>
        >>> # set a page's text to the empty string
        >>> zc.TestPage.edit(text='')
        >>> zc.TestPage.src()
        ''
        >>>
        >>> # add a file to a page
        >>> import OFS.Image
        >>> file = OFS.Image.Pdata('test file data')
        >>> file.filename = 'edittestfile'
        >>> zc.REQUEST.file = file
        >>> zc.TestPage.edit(REQUEST=zc.REQUEST)
        >>> # the new file should exist
        >>> assert hasattr(zc,'edittestfile')
        >>> # with the right data
        >>> str(zc['edittestfile'])
        'test file data'
        >>> # and a link should have been added to the page
        >>> zc.TestPage.src()
        '\\012\\012!<a href="edittestfile">edittestfile</a>\\012'
        >>> # cleanup - temporary I think, because of double testing
        >>> zc.manage_delObjects(['edittestfile']) 
        >>>
        >>> # ditto, with an image
        >>> zc.REQUEST.file.filename = 'edittestimage.jpg'
        >>> zc.TestPage.edit(REQUEST=zc.REQUEST)
        >>> assert hasattr(zc,'edittestimage.jpg')
        >>> zc['edittestimage.jpg'].content_type
        'image/jpeg'
        >>> zc.TestPage.src()
        '\\012\\012!<a href="edittestfile">edittestfile</a>\\012\\012\\012!<img src="edittestimage.jpg">\\012'
        >>> # cleanup
        >>> zc.manage_delObjects(['edittestimage.jpg']) 
        >>>
        >>> # images should not be inlined if dontinline is set
        >>> zc.REQUEST.file.filename = 'edittestimage.png'
        >>> zc.REQUEST.dontinline = 1
        >>> zc.TestPage.edit(REQUEST=zc.REQUEST)
        >>> zc.TestPage.src()
        '\\012\\012!<a href="edittestfile">edittestfile</a>\\012\\012\\012!<img src="edittestimage.jpg">\\012\\012\\012!<a href="edittestimage.png">edittestimage.png</a>\\012'
        >>> # cleanup
        >>> zc.manage_delObjects(['edittestimage.png']) 
        >>> zc.REQUEST.dontinline = None
        >>>
        >>> # a file with blank filename should be ignored
        >>> zc.REQUEST.file.filename = ''
        >>> old = zc.TestPage.src()
        >>> zc.TestPage.edit(REQUEST=zc.REQUEST)
        >>> assert zc.TestPage.src() == old


        DeleteMe
        (this feature was actually coded test-first)
        How should this work ? Do the simplest thing. When we see a first
        line beginning with "DeleteMe":
        - move to recycle_bin
        - redirect to first parent or default page

        create a test page
        >>> zc.manage_addProduct['ZWiki'].manage_addZWikiPage(\
                'TestPageA', file='not much')
        ''

        deleteme's not at the beginning shouldn't do anything
        >>> zc.TestPageA.edit(text=zc.TestPageA.src()+'\\nDeleteMe')
        >>> zc.TestPageA.src()
        'not much\\012DeleteMe'

        XXX deleteme at the beginning will send it to recycle_bin/:
        >>> zc.TestPageA.edit(text='DeleteMe, with comments\\n'+zc.TestPageA.src())
        >>> assert not hasattr(zc,'TestPageA')
        >>> zc.recycle_bin.TestPageA.src()
        'DeleteMe, with comments\\012not much\\012DeleteMe'


        Username stamping (old)
        problem: how do we manipulate the authenticated user and
        still run tests ?

        #if we are authenticated, our auth. name should be recorded
        #>>> zc.TestPage.username = '-'
        #>>> zc.REQUEST.REMOTE_ADDR = '1.2.3.4'
        #>>> zc.REQUEST.cookies['zwiki_username'] = 'TESTUSERNAME'
        #>>> zc.TestPage.edit(text=zc.TestPage.src()+'.',REQUEST=zc.REQUEST)
        #>>> assert zc.TestPage.username == zc.REQUEST.AUTHENTICATED_USER.getUserName()

        #if a cookie is not present, IP address should be used
        #>>> zc.REQUEST.REMOTE_ADDR = '1.2.3.4'
        #>>> zc.TestPage.edit(text=zc.TestPage.src()+'.',REQUEST=zc.REQUEST)
        #>>> zc.TestPage.username
        #'1.2.3.4'

        #NB username will not be stamped unless something is actually
        #changed
        #>>> zc.REQUEST.REMOTE_ADDR = '5.6.7.8'
        #>>> zc.TestPage.edit(text=zc.TestPage.src(),
        #...                   type=zc.TestPage.page_type,REQUEST=zc.REQUEST)
        #>>> zc.TestPage.username
        #'1.2.3.4'

        #if the zwiki_username cookie is present, it should be used
        #>>> zc.REQUEST.REMOTE_ADDR = '1.2.3.4'
        #>>> zc.REQUEST.cookies['zwiki_username'] = 'TESTUSERNAME'
        #>>> zc.TestPage.edit(text=zc.TestPage.src()+'.',REQUEST=zc.REQUEST)
        #>>> zc.TestPage.username
        #'TESTUSERNAME'

        #if the cookie is present but blank, IP address should be used
        #>>> zc.REQUEST.REMOTE_ADDR = '1.2.3.4'
        #>>> zc.REQUEST.cookies['zwiki_username'] = ''
        #>>> zc.TestPage.edit(text=zc.TestPage.src()+'.',REQUEST=zc.REQUEST)
        #>>> zc.TestPage.username
        #'1.2.3.4'

        """
        
        #self._validateProxy(REQUEST)   # XXX correct ? don't think so
                                        # do zwiki pages obey proxy roles ?
        checkPermission = getSecurityManager().checkPermission

        # what page are we changing ?

        # NB web browsers want spaces "like%20so", zope wants ids "like so"
        # XXX let's see what this breaks
        if page: page = unquote(page)
        
        if page is None:
            # changing the current page
            p = self
        elif hasattr(self.aq_parent, page):
            # changing another specified page
            p = getattr(self.aq_parent, page)
        else:
            # creating a new page
            if not checkPermission(ZWikiPermissions.Add,self.aq_parent):
                raise 'Unauthorized', (
                    'You are not authorized to add ZWiki Pages here.')
            return self._create_page(page,text,type,title,REQUEST)

        # ok, changing p:

        # first, check for an edit conflict
        if self.checkEditTimeStamp(timeStamp, REQUEST):
            return MessageDialog(
                title='Edit conflict!',
                message="""
                <b>Edit conflict!</b>
                <p>
                Someone else has saved this page while you were
                editing. To resolve the conflict, do this:
                <ol>
                <li>Click your browser's back button
                <li>Copy your recent edits to the clipboard
                <li>Click your browser's refresh button
                <li>Paste in your edits again, being mindful of the latest changes
                <li>Click the Change button again.
                </ol>
                or,
                <p>
                To discard your recent edit and start again, click OK.
                """,
                action=self.page_url()+'/editform')

        # 1. change page type ?
        if (type is not None and
            type != p.page_type):

            if not checkPermission(ZWikiPermissions.ChangeType,p):
                raise 'Unauthorized', (
                    'You are not authorized to change this ZWiki Page\'s type.')

            p.page_type = type
            p._set_last_editor(REQUEST)

        # 2. change page text ?
        if (text is not None and
            self._text_cleanups(text) != p.xread()):

            if (not
                (checkPermission(ZWikiPermissions.Change, p) or
                 (checkPermission(ZWikiPermissions.Append, p)
                  and find(self._text_cleanups(text),p.xread()) == 0))):
                raise 'Unauthorized', (
                    'You are not authorized to edit this ZWiki Page.')

            # handle DeleteMe special case ?
            if re.match('(?m)^DeleteMe', text):
            
                if not checkPermission(ZWikiPermissions.Delete, p):
                    raise 'Unauthorized', (
                        'You are not authorized to send this ZWiki Page to the recycle bin.')
                
                # change text
                p._set_text(text,REQUEST)
                p._set_last_editor(REQUEST)

                # XXX move to p.delete():

                # create recycle_bin folder if needed
                f = self.aq_parent
                if not hasattr(f,'recycle_bin'):
                    f.manage_addFolder('recycle_bin', 'deleted wiki pages')

                # move page there
                id=p.id() # XXX use getId() when available
                cb=f.manage_cutObjects(id)
                f.recycle_bin.manage_pasteObjects(cb)

                # XXX update catalog

                # redirect to first parent, or the front page                
                if REQUEST is not None:
                    try:
                        destpage=f.recycle_bin[id].parents[0] # XXX parent code
                        assert(hasattr(f,destpage))
                    except:
                        destpage=''
                    REQUEST.RESPONSE.redirect(
                        self.wiki_url()+'/'+quote(destpage))
                    
                return

            # change text
            p._set_text(text,REQUEST)
            p._set_last_editor(REQUEST)

        # 3. upload a file or image ?
        if (REQUEST and
            hasattr(REQUEST,'file') and
            hasattr(REQUEST.file,'filename') and
            REQUEST.file.filename):     # XXX do something

            # CHECK permissions
            if not (checkPermission(ZWikiPermissions.Change, p) or
                    checkPermission(ZWikiPermissions.Append, p)):
                raise 'Unauthorized', (
                    'You are not authorized to add a link on this ZWiki Page.')

            if hasattr(p,'uploads'):
                uploaddir = p.uploads
            else:
                uploaddir = self.aq_parent
            if not (checkPermission(ZWikiPermissions.Upload,uploaddir)):# or
                    #checkPermission(ZWikiPermissions.UploadSmallFiles,
                    #                self.aq_parent)):
                raise 'Unauthorized', (
                    'You are not authorized to add files here.')

            # XXX how to get file's size ?
            #if (len(REQUEST.file) > LARGE_FILE_SIZE and
            #    not checkPermission(ZWikiPermissions.Upload,
            #                        self.aq_parent)):
            #    raise 'Unauthorized', (
            #        'You are not authorized to add files larger than ' + \
            #        LARGE_FILE_SIZE + ' here.')

            # create file or image object
            file_id, content_type, size = \
                    self._create_file(REQUEST.file,
                                      title=REQUEST.get('title', ''),
                                      REQUEST=REQUEST)

            # and add a link to it
            if file_id:
                if hasattr(p,'uploads'):
                    filepath = 'uploads/'
                else:
                    filepath = ''

                # we have the size now - too late for permissions but
                # at least we can use it to help decide whether to
                # inline an image
                if content_type[0:5] == 'image' and \
                   not (hasattr(REQUEST,'dontinline') and REQUEST.dontinline) and \
                   size <= LARGE_FILE_SIZE :
                    linktxt = '\n\n!<img src="%s%s">\n' % (filepath,file_id)
                else:
                    linktxt = '\n\n!<a href="%s%s">%s</a>\n' % (filepath,file_id,file_id)

                p._set_text(p.src()+linktxt,REQUEST,self.aq_parent)
                p._set_last_editor(REQUEST)

        # update this page's catalog entry
        #p.reindex_object()

        # redirect browser, if needed
        if REQUEST is not None:
            try: u=p.DestinationURL()   # XXX lookie here.. what can we do with this
            except: u=REQUEST['URL1']
            REQUEST.RESPONSE.redirect(u)

    def _create_page(self, page, text=None, type=None, title='', REQUEST=None):
        """
        Create a new page and display it. A helper method called by edit().

        TESTS:

        >>> # create a blank page
        >>>
        >>> zc.TestPage._create_page('TestPage1',text='')
        >>> assert hasattr(zc,'TestPage1')
        >>> zc.TestPage1.text()
        ''
        >>>
        >>> # create a classicwiki page with some text
        >>>
        >>> zc.TestPage._create_page('TestPage2',text='test page data',type='classicwiki')
        >>> assert hasattr(zc,'TestPage1')
        >>> zc.TestPage2.src()
        'test page data'
        >>> zc.TestPage2.page_type
        'classicwiki'
        >>>
        >>> # add a file while creating a page
        >>>
        >>> import OFS.Image
        >>> file = OFS.Image.Pdata('test file data')
        >>> file.filename = 'test_file'
        >>> zc.REQUEST.file = file
        >>> zc.TestPage._create_page('TestPage3',text='test page data',REQUEST=zc.REQUEST)
        >>> # the new file should exist
        >>> assert hasattr(zc,'test_file')
        >>> # with the right data
        >>> str(zc['test_file'])
        'test file data'
        >>> # and a link should have been added to the new wiki page
        >>> zc.TestPage3.src()
        'test page data\\012\\012!<a href="test_file">test_file</a>\\012'
        >>>
        >>> # ditto, with an image
        >>>
        >>> file.filename = 'test_image.gif'
        >>> zc.REQUEST.file = file
        >>> zc.TestPage1._create_page('TestPage4',text='test page data',REQUEST=zc.REQUEST)
        >>> assert hasattr(zc,'test_image.gif')
        >>> zc['test_image.gif'].content_type
        'image/gif'
        >>> zc.TestPage4.src()
        'test page data\\012\\012!<img src="test_image.gif">\\012'
        >>>
        >>> # images should not be inlined if dontinline is set
        >>>
        >>> file.filename = 'test_image.JPG'
        >>> zc.REQUEST.dontinline = 1
        >>> zc.TestPage1._create_page('TestPage5',text='',REQUEST=zc.REQUEST)
        >>> zc.TestPage5.src()
        '\\012\\012!<a href="test_image.JPG">test_image.JPG</a>\\012'
        >>> # cleanup
        >>> zc.REQUEST.dontinline = None

        """
        # make a new (blank) page object
        ob = ZWikiPage(source_string='', __name__=page)
        ob.title = title
        ob.parents = [self.id()] # XXX parent code
        ob._set_last_editor(REQUEST)

        # inherit type from the previous (parent) page if not specified
        # or overridden via folder attribute
        if hasattr(self.aq_parent,'standard_page_type'):
            ob.page_type = self.aq_parent.standard_page_type
        elif type:
            ob.page_type = type
        else:
            ob.page_type = self.page_type


        # situate it in the parent folder
        # NB newid might not be page, eg if an object with that id
        # already exists it would be "copy of page"
        newid = self.aq_parent._setObject(page,ob)

        # 2.2-specific: the new page object is owned by the current
        # authenticated user, if any; not desirable for executable content.
        # Remove any such ownership so that the page will acquire it's
        # owner from the parent folder.
        ob._deleteOwnershipAfterAdd()

        # choose initial page text
        if text is not None:
            t = text
        elif hasattr(self.aq_parent, 'standard_wiki_page'):
            if callable(self.aq_parent.standard_wiki_page):
                t = self.aq_parent.standard_wiki_page(self,REQUEST)
            else:
                t = HTML(self.aq_parent.standard_wiki_page)
        else:
            t = default_wiki_page(self,REQUEST)

        # now, set the text as edit() would, to do any cleanups and
        # to roll this all back if there's a DTML error
        # pass parent as an arg - see _set_text()
        ob._set_text(t,REQUEST,self.aq_parent)

        # also, if a file was submitted, create a suitable file or
        # image object and link or inline it
        if REQUEST and hasattr(REQUEST,'file'):
            file = REQUEST.file
            if file:
                id, content_type, size = \
                    self._create_file(file,
                                      title=REQUEST.get('title', ''),
                                      REQUEST=REQUEST)
                if id:
                    if hasattr(self,'uploads'):
                        filepath = 'uploads/'
                    else:
                        filepath = ''
                    if content_type[0:5] == 'image' and \
                       not (hasattr(REQUEST,'dontinline') and REQUEST.dontinline) and \
                       size <= LARGE_FILE_SIZE :
                        link = '\n\n!<img src="%s%s">\n' % (filepath,id)
                    else:
                        link = '\n\n!<a href="%s%s">%s</a>\n' % (filepath,id,id)
                    ob._set_text(t+link,REQUEST,self.aq_parent)

        # catalog code
        # add it to the catalog
        # need to do this sooner
        # can't use ob.index_object(), because ob doesn't have the
        # zpublisher/acquisition wrapper I think
        #if hasattr(self, self.default_catalog):
        #    url = join(split(self.url(),'/')[:-1],'/') + '/' + newid   # !!??
        #    getattr(self, self.default_catalog).catalog_object(ob, url)

        if REQUEST is not None:
            # redirect browser to the new page
            # all these redirects.. better/safer way ?
            u=REQUEST['URL2']
            REQUEST.RESPONSE.redirect(u + '/' + quote(newid))

    def _create_file(self,file,title='',REQUEST=None):
        # Lifted from OFS/Image.py:File:manage_addFile
        # Lifted from WikiForNow
        """
        Add a new File or Image object, depending on file's filename
        suffix. Returns a tuple containing the new id, content type &
        size, or (None,None,None).

        TESTS:

        >>> import OFS.Image
        >>> file = OFS.Image.Pdata('test file data')
        >>> 
        >>> # our test page/folder should initially have no uploads attr.
        >>> self = zc.TestPage
        >>> #assert not hasattr(self,'uploads')
        >>>
        >>> # calling with an unnamed file should do nothing much
        >>> zc.TestPage._create_file(file)
        (None, None, None)
        >>>
        >>> # ditto for a blank filename
        >>> file.filename = ''
        >>> zc.TestPage._create_file(file)
        (None, None, None)
        >>>
        >>> # here, a file object of unknown type should be created
        >>> name = 'testfile'
        >>> file.filename = name
        >>> id, content_type,size = zc.TestPage._create_file(file)
        >>> str(zc[id])
        'test file data'
        >>> content_type 
        'application/octet-stream'
        >>> # was 'text/x-unknown-content-type'
        >>> size
        14
        >>>
        >>> # a text file
        >>> name = 'testfile.txt'
        >>> file.filename = name
        >>> id, content_type,size = zc.TestPage._create_file(file)
        >>> str(zc[id])
        'test file data'
        >>> content_type
        'text/plain'
        >>>
        >>> # an image
        >>> name = 'testfile.gif'
        >>> file.filename = name
        >>> id, content_type,size = zc.TestPage._create_file(file)
        >>> str(zc[id])[-47:] # evaluating an Image gives its html tag
        '/testfile.gif" alt="testfile.gif" border="0" />'
        >>> content_type
        'image/gif'

        """

        # set id & title from filename
        title=str(title)
        id, title = OFS.Image.cookId('', title, file)
        if not id:
            return None, None, None

        # find out where to store files - in the 'uploads'
        # folder if defined, otherwise the wiki folder
        # NB a page might override this with a property
        if hasattr(self,'uploads'):
            folder = self.uploads
        else:
            folder = self.aq_parent

        # First, we create the file or image object without data:
        # XXX SM handle existing object ?
        if guess_content_type(file.filename)[0][0:5] == 'image':
            folder._setObject(id, OFS.Image.Image(id,title,''))
        else:
            folder._setObject(id, OFS.Image.File(id,title,''))

        # Now we "upload" the data.  By doing this in two steps, we
        # can use a database trick to make the upload more efficient.
        folder._getOb(id).manage_upload(file)

        return id, folder._getOb(id).content_type, folder._getOb(id).getSize()

    # NB when called from _create_page() we have to pass in parent
    # here because the change that would make self.parent
    # work has not yet been commited to the db
    # XXX check for correct use of parent by callers
    # XXX parent arg no longer used ?
    def _set_text(self, text='', REQUEST=None, parent=None):
        """change the page text, with cleanups and perhaps DTML validation
        """
        t = self._text_cleanups(text)

        # for DTML page types, execute the DTML to catch problems -
        # zope will magically roll back this whole transaction and the
        # user will get an appropriate error
        if re.search(r'(?i)dtml',self.page_type):

            # XXX see render_structuredtextdtml
            t = '<!--antidecapitationkludge-->\n\n' + t

            self.munge(t)
            # XXX problem: dtml in a newly-created page will give unauthorized
            # a commit doesn't seem to help
            #get_transaction().commit()
            # so I'm going to skip dtml validation in this case
            # a person entering dtml in a new page knows what they're doing
            # and should be able to go to /editform if needed
            # XXX refactoring bug - now we don't do any dtml validation at all?
            #DTMLDocument.__call__(self,self,REQUEST,REQUEST.RESPONSE)
        else:
            self.raw = t

    def _text_cleanups(self, t):
        """do some cleanup of a page's new text
        """
        # strip any browser-appended ^M's
        t = re.sub('\r\n', '\n', t)

        # convert international characters to HTML entities for safekeeping
        for c,e in intl_char_entities:
            t = re.sub(c, e, t)

        # here's the place to strip out any disallowed html/scripting elements 
        if DISABLE_JAVASCRIPT:
            # XXX needs work
            # don't match <input... name="ZPythonScriptHTML_editAction:method">
            #t = re.sub(r'(?i)<([^d>]*script[^>]*)>',r'&lt;disabled \1&gt;',t)
            t = re.sub(r'(?i)<([^>\w]*script[^>]*)>',r'&lt;disabled \1&gt;',t)

        return t

    def _set_last_editor(self, REQUEST=None):
        if REQUEST:
            self.last_editor_ip = REQUEST.REMOTE_ADDR
            self.last_editor = self.zwiki_username_or_ip()   

    # another utility method that can be handy in scripts
    zwiki_username_or_ip__roles__ = None
    def zwiki_username_or_ip(self, REQUEST=None):
        """
        search REQUEST for an authenticated member or a zwiki_username
        cookie
        """
        # XXX doesn't work - why is self.REQUEST a string ?
        #if not REQUEST and hasattr(self,'REQUEST'):
        #    REQUEST = self.REQUEST
        username = None
        if REQUEST:
            user = REQUEST.AUTHENTICATED_USER
            username = user.getUserName()
            if not username or str(user.acl_users._nobody) == username:
                if hasattr(REQUEST, 'cookies') and \
                REQUEST.cookies.has_key('zwiki_username') and \
                REQUEST.cookies['zwiki_username']:
                    username = REQUEST.cookies['zwiki_username']
                else:
                    username = REQUEST.REMOTE_ADDR
        return username or ''

    # jim's safety belts for warning of http & ftp edit conflicts
    def manage_FTPget(self):
        "Get source for FTP download"
        return "Wiki-Safetybelt: %s\n\n%s" % (
            self.editTimestamp(), self.xread())

    # should this do what _set_text() does ?
    def PUT(self, REQUEST, RESPONSE):
        """Handle HTTP PUT requests."""
        self.dav__init(REQUEST, RESPONSE)
        body=REQUEST.get('BODY', '')
        self._validateProxy(REQUEST)

        m=re.match(r'Wiki-Safetybelt: ([0-9]+[.][0-9]+)[\n][\n]?', body)
        if m:
            body = body[m.span()[1]:]            
            if self.checkEditTimeStamp(m.group(1), REQUEST):
                RESPONSE.setStatus(404) #XXX what should this be
                return RESPONSE
                
        self.munge(body)
        self._set_last_editor(REQUEST)

        RESPONSE.setStatus(204)
        return RESPONSE

    # edit conflict checking
    # how can we be smarter about edit conflicts, eg in the common
    # case where you back up in your browser, edit and click the
    # change button again ? Ignore the timestamp if our username & IP
    # match the last editor's ?
    #
    # XXX should we consider the append form in this ?
    #
    # prior musings which may be a load of codswallop by now:
    # What if
    # - we are behind a proxy so all ip's are the same ?
    # - several people use the same cookie-based username ?
    # - people use the same cookie-name as an existing member name ?
    # - noone is using usernames ?
    # in the proxy case, we'll usually get a username mismatch as long
    # as we exclude "anonymous" and ''. But this would require
    # anonymous users to configure a username or experience the old
    # strict behaviour, which is a drag.  By default we'll assume no
    # proxy and let these users by on the strength of their ip.
    #
    # so 4 strategies so far:
    #
    # 0. no conflict checking
    #
    # 1. strict - based only on timestamp. (The behaviour of previous
    # releases. Safest but obstructs backtrack & re-edit).
    #
    # 2. semi-careful - include username & ip address in the
    # "timestamp". If your (non-anonymous) username and ip match, a
    # time mismatch will be ignored.  (If a proxy is present there
    # will be no conflict checking amongst users with the same member-
    # or cookie-name. Anonymous users will experience strict checking
    # until they configure a username.)
    #
    # 3. relaxed - if your username & ip match, a time mismatch will
    # be ignored. (The new default. If a proxy is present there will
    # be no conflict checking amongst anonymous users).
    #
    # Started making this an option but it's too much hassle

    def editTimestamp(self):
        return str(self._p_mtime)
    
    def checkEditTimeStamp(self, timeStamp, REQUEST):
        username = self.zwiki_username_or_ip()
        if (timeStamp is not None and
            timeStamp != self.editTimestamp() and
            (not hasattr(self,'last_editor') or
             not hasattr(self,'last_editor_ip') or
             username != self.last_editor or
             REQUEST.REMOTE_ADDR != self.last_editor_ip)):
            # XXX how can we do something more friendly here
            #raise 'EditingConflict', (
            #    '''Someone has edited this page since you loaded the
            #    page for editing.<p>
            #    Try editing the page again.
            #    ''')
            return 1
        else:
            return 0


    ########################################
    # misc methods

    def backlinks(self, REQUEST=None):
        """
        Display a default backlinks page. May be overridden by a DTML
        method of the same name.
        """
        if hasattr(self.aq_parent, 'backlinks'):
            return self.aq_parent.backlinks(self,REQUEST)
        else:
            return default_backlinks(self,REQUEST)

    # backwards compatibility for dtml which tries to display the old
    # username property. Don't assume the page has auto-upgraded yet.
    def username(self):
        for p in self._properties:
            if p['id'] == 'username':
                return username

        if hasattr(self,'last_editor') and self.last_editor:
            return self.last_editor

        if hasattr(self,'last_editor_ip') and self.last_editor_ip:
            return self.last_editor_ip

        return ''
        
    # actually this is useful to keep around as
    last_editor_or_ip = username
    last_editor_or_ip__roles__ = None


    # methods for reliable access to page & wiki url
    # these have been troublesome; they need to work with, without
    # virtual hosting. Keep old versions around to facilitate debugging.
    # Need tests.
    # XXX cleanup
    
#    def page_url(self):
#        """return the url path for this wiki page"""
#        return self.wiki_url() + '/' + self.id()
#
#    def wiki_url(self):
#        """return the base url path for this wiki"""
#        return self.aq_inner.aq_parent.absolute_url()
#
#    # backwards compatibility
#    wiki_page_url = page_url
#    wiki_page_url__roles__ = None
#    wiki_base_url = wiki_url
#    wiki_base_url__roles__ = None
#
# we'll proceed with the celebratory rename when this is settled

    def wiki_page_url(self):
        """return the url path for the current wiki page"""
        return self.wiki_page_url7()

    def wiki_base_url(self):
        """return the base url path for the current wiki"""
        return self.wiki_base_url7()

    # forwards compatibility
    # can you tell I'm confused ?
    page_url = wiki_page_url
    page_url__roles__ = None
    wiki_url = wiki_base_url
    wiki_url__roles__ = None

    # XXX TESTING - see http://zwiki.org/VirtualHostingSummary
    def wiki_page_url1(self):
        """return the url path for the current wiki page"""
        o = self
        url = []
        while hasattr(o,'id'):
            url.insert(0,absattr(o.id))
            o = getattr(o,'aq_parent', None)
        return quote('/' + join(url[1:],'/'))

    def wiki_base_url1(self):
        """return the base url path for the current wiki"""
        # somewhere in cvs
        return '?'

    def wiki_page_url2(self):
        """return the url path for the current wiki page"""
        return quote('/' + self.absolute_url(relative=1))

    def wiki_base_url2(self):
        """return the base url path for the current wiki"""
        return quote('/' + self.aq_inner.aq_parent.absolute_url(relative=1))

    def wiki_page_url3(self):
        """return the url path for the current wiki page"""
        return self.REQUEST['BASE1'] + '/' + self.absolute_url(relative=1)

    def wiki_base_url3(self):
        """return the base url path for the current wiki"""
        return self.REQUEST['BASE1'] + '/' + self.aq_inner.aq_parent.absolute_url(relative=1)

    def wiki_page_url4(self):
        """return the url path for the current wiki page"""
        return self.absolute_url(relative=0)

    def wiki_base_url4(self):
        """return the base url path for the current wiki"""
        return self.aq_inner.aq_parent.absolute_url(relative=0)

    def wiki_page_url5(self):
        """return the url path for the current wiki page"""
        return self.absolute_url()

    def wiki_base_url5(self):
        """return the base url path for the current wiki"""
        return self.aq_inner.aq_parent.absolute_url()

    def wiki_page_url6(self):
        """return the url path for the current wiki page"""
        return '/' + self.absolute_url(relative=1)

    def wiki_base_url6(self):
        """return the base url path for the current wiki"""
        return '/' + self.aq_inner._my_folder().absolute_url(relative=1)

    def _my_folder(self):
        """Obtain parent folder, avoiding potential acquisition recursion."""
        got = parent = self.aq_parent
        # Handle bizarred degenerate case - darnit, i've forgotten now where
        # i've seen this occur!
        while (got
               and hasattr(got, 'aq_parent')
               and got.aq_parent
               and (got is not got.aq_parent)
               and (aq_base(got) is aq_base(got.aq_parent))):
            parent = got
            got = got.aq_parent
        return parent

    def wiki_page_url7(self):
        """return the url path for the current wiki page"""
        return self.wiki_base_url() + '/' + quote(self.id())

    def wiki_base_url7(self):
        """return the base url path for the current wiki"""
        return self.aq_inner.aq_parent.absolute_url()


# end of ZWikiPage class


########################################
# rendering helper functions

def thunk_substituter(func, text, allowed):
    """Return a function which takes one arg and passes it with other args
    to passed-in func.

    thunk_substituter passes in the value of it's parameter, 'allowed', and a
    dictionary {'lastend': int, 'inpre': bool, 'intag': bool}.

    This is for use in a re.sub situation, to get the 'allowed' parameter and
    the state dict into the callback.

    (The technical term really is "thunk".  Honest.-)"""
    state = {'lastend': 0, 'inpre': 0, 'incode': 0, 'intag': 0}
    return lambda arg, func=func, allowed=allowed, text=text, state=state: (
        func(arg, allowed, state, text))

def within_literal(upto, after, state, text,
                   rfind=rfind, lower=lower):
    """Check text from state['lastend'] to upto for literal context:

    - Within an enclosing '<pre>' preformatted region '</pre>'
    - Within an enclosing '<code>' code fragment '</code>'
    - Within a tag '<' body '>'

    We also update the state dict accordingly."""
    # XXX This breaks on badly nested angle brackets and <pre></pre>, etc.
    lastend, inpre, intag = state['lastend'], state['inpre'], state['intag']
    lastend = state['lastend']
    inpre, incode, intag = state['inpre'], state['incode'], state['intag']
    newintag = newincode = newinpre = 0
    text = lower(text)

    # Check whether '<pre>' is currently (possibly, still) prevailing.
    opening = rfind(text, '<pre>', lastend, upto)
    if (opening != -1) or inpre:
        if opening != -1: opening = opening + 4
        else: opening = lastend
        if -1 == rfind(text, '</pre>', opening, upto):
            newinpre = 1
    state['inpre'] = newinpre

    # Check whether '<code>' is currently (possibly, still) prevailing.
    opening = rfind(text, '<code>', lastend, upto)
    if (opening != -1) or incode:
        if opening != -1: opening = opening + 5
        # We must already be incode, start at beginning of this segment:
        else: opening = lastend
        if -1 == rfind(text, '</code>', opening, upto):
            newincode = 1
    state['incode'] = newincode

    # Determine whether we're (possibly, still) within a tag.
    opening = rfind(text, '<', lastend, upto)
    if (opening != -1) or intag:
        # May also be intag - either way, we skip past last <tag>:
        if opening != -1: opening = opening + 1
        # We must already be intag, start at beginning of this segment:
        else: opening = lastend
        if -1 == rfind(text, '>', opening, upto):
            newintag = 1
    state['intag'] = newintag

    state['lastend'] = after
    return newinpre or newincode or newintag

# exact reverse of DT_Util.html_quote
def html_unquote(v, name='(Unknown name)', md={},
                 character_entities=(
                       (('&amp;'),    '&'),
                       (('&lt;'),    '<' ),
                       (('&gt;'),    '>' ),
                       (('&lt;'), '\213' ),
                       (('&gt;'), '\233' ),
                       (('&quot;'),    '"'))): #"
        text=str(v)
        for re,name in character_entities:
            if find(text, re) >= 0: text=join(split(text,re),name)
        return text


# ensure that doctest will test all ZWikiPage's private methods
# XXX but this causes public methods to be tested twice, fix
__test__ = { "ZWikiPage": ZWikiPage, }


# page & wiki creation
from ZWikiAdd import *


